"""pypyr step that merges the input mappings into context.

Whereas contextcopy and contextsetf overwrites values that are in context
already, contextmerge merges its input into context, preserving the existing
hierarchy while just updating the values where specified in the contextmerge
input.

Applies string interpolation as it merges. String interpolation applies to keys
and values.
"""
import logging

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Merge hierarchy into context with substitutions.

    context is a dictionary or dictionary-like.
    context['contextMerge'] must exist. It's a dictionary.

    Will iterate context['contextMerge'] and save the values as new keys to the
    context where they exist already, and add these as new values where they
    don't already exist. While it's doing so, it will leave unspecified values
    in the existing hierarchy untouched.

    List merging is purely additive, with no checks for uniqueness or already
    existing list items. E.g context [0,1,2] with contextMerge=[2,3,4]
    will result in [0,1,2,2,3,4]

    Keep this in mind especially where complex types like
    dicts nest inside a list - a merge will always add a new dict list item,
    not merge it into whatever dicts might exist on the list already.

    For example, say input context is:
        key1: value1
        key2: value2
        key3:
            k31: value31
            k32: value32
        contextMerge:
            key2: 'aaa_{key1}_zzz'
            key3:
                k33: value33
            key4: 'bbb_{key2}_yyy'

    This will result in return context:
        key1: value1
        key2: aaa_value1_zzz
        key3:
            k31: value31
            k32: value32
            k33: value33
        key4: bbb_value2_yyy
    """
    logger.debug("started")
    context.assert_key_has_value(key='contextMerge', caller=__name__)

    context.merge(context['contextMerge'])

    logger.info("merged %d context items.", len(context['contextMerge']))

    logger.debug("done")



"""pypyr step that gets paths from a glob."""
import logging
from pypyr.errors import KeyInContextHasNoValueError
import pypyr.utils.filesystem

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Get paths from a glob input.

    Do note that this will returns files AND directories that match the glob.

    No tilde expansion is done, but *, ?, and character ranges expressed with
    [] will be correctly matched.

    Escape all special characters ('?', '*' and '['). For a literal match, wrap
    the meta-characters in brackets. For example, '[?]' matches the character
    '?'.

    If passing in an iterable of paths, will expand matches for each path in
    the iterable. The function will return all the matches for each path
    glob expression combined into a single list.

    If no matches found, writes empty list [] to globOut.

    Args:
        context: pypyr.context.Context. Mandatory.
                 The following context key is mandatory:
                - glob. str or list. Single path, or list of paths.

    All inputs support pypyr formatting expressions.

    Returns:
        None. updates context ['globOut'] with list of resolved paths as
              strings.

    """
    logger.debug("started")
    context.assert_key_has_value(key='glob', caller=__name__)

    paths = context.get_formatted('glob')
    if isinstance(paths, list):
        if not paths or any(not p for p in paths):
            raise KeyInContextHasNoValueError("The glob list has an empty str")
        in_count = len(paths)
    else:
        if not paths:
            raise KeyInContextHasNoValueError("The glob path is an empty str")
        in_count = 1

    context['globOut'] = pypyr.utils.filesystem.get_glob(paths)

    logger.info("glob checked %s globs and saved "
                "%s paths to globOut", in_count, len(context['globOut']))
    logger.debug("done")



"""Context parser that returns a list from input arguments.

Takes input args (i.e separated by spaces on cli) and returns a list named
argList.

So a string like this "ham eggs bacon", will yield context:
{ 'argList': ['ham', 'eggs', 'bacon']}
"""
import logging

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)


def get_parsed_context(args):
    """Parse input context args and returns context as dictionary."""
    logger.debug("starting")
    if not args:
        logger.debug("pipeline invoked without context arg set. For "
                     "this list parser you're looking for something like:\n"
                     "pypyr pipelinename spam eggs\n"
                     "OR: pypyr pipelinename spam."
                     )
        return {'argList': None}

    # the list that's parsed from the input args is named argList
    return dict({'argList': args})



"""pypyr step that executes a string as python.

Uses python's exec() to evaluate and execute arbitrary python code.
"""
import logging

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Execute dynamic python code.

    Context is a dictionary or dictionary-like.
    Context must contain key 'pycode'
    Will exec context['pycode'] as dynamically interpreted python statements.

    context is mandatory. When you execute the pipeline, it should look
    something like this:
        pipeline-runner [name here] 'pycode=print(1+1)'.
    """
    logger.debug("started")
    context.assert_key_has_value(key='pycode', caller=__name__)

    logger.debug("Executing python string: %s", context['pycode'])
    locals_dictionary = locals()
    exec(context['pycode'], globals(), locals_dictionary)

    # It looks like this dance might be unnecessary in python 3.6
    logger.debug("looking for context update in exec")
    exec_context = locals_dictionary['context']
    context.update(exec_context)
    logger.debug("exec output context merged with pipeline context")

    logger.debug("done")



"""pypyr step that runs another pipeline from within the current pipeline."""
import logging
from pypyr.context import Context
from pypyr.errors import (ContextError,
                          ControlOfFlowInstruction,
                          KeyInContextHasNoValueError,
                          KeyNotInContextError,
                          Stop)
import pypyr.pipelinerunner as pipelinerunner

# logger means the log level will be set correctly
logger = logging.getLogger(__name__)


def run_step(context):
    """Run another pipeline from this step.

    The parent pipeline is the current, executing pipeline. The invoked, or
    child pipeline is the pipeline you are calling from this step.

    Args:
        context: dictionary-like pypyr.context.Context. context is mandatory.
                 Uses the following context keys in context:
            - pype
                - name. mandatory. str. Name of pipeline to execute. This
                  {name}.yaml must exist in the working directory/pipelines
                  dir.
                - args. optional. dict. Create the context of the called
                  pipeline with these keys & values. If args specified,
                  will not pass the parent context unless you explicitly set
                  useParentContext = True. If you do set useParentContext=True,
                  will write args into the parent context.
                - out. optional. str or dict or list. If you set args or
                  useParentContext=False, the values in out will be saved from
                  child pipeline's fresh context into the parent content upon
                  completion of the child pipeline. Pass a string for a single
                  key to grab from child context, a list of string for a list
                  of keys to grab from child context, or a dict where you map
                  'parent-key-name': 'child-key-name'.
                - pipeArg. string. optional. String to pass to the
                  context_parser - the equivalent to context arg on the
                  pypyr cli. Only used if skipParse==False.
                - raiseError. bool. optional. Defaults to True. If False, log,
                  but swallow any errors that happen during the invoked
                  pipeline execution. Swallowing means that the current/parent
                  pipeline will carry on with the next step even if an error
                  occurs in the invoked pipeline.
                - skipParse. bool. optional. Defaults to True. skip the
                  context_parser on the invoked pipeline.
                - useParentContext. optional. bool. Defaults to True. Pass the
                  current (i.e parent) pipeline context to the invoked (child)
                  pipeline.
                - loader: str. optional. Absolute name of pipeline loader
                  module. If not specified will use
                  pypyr.pypeloaders.fileloader.
                - groups. list of str, or str. optional. Step-Groups to run in
                  pipeline. If you specify a str, will convert it to a single
                  entry list for you.
                - success. str. optional. Step-Group to run on successful
                  pipeline completion.
                - failure. str. optional. Step-Group to run on pipeline error.

    If none of groups, success & failure specified, will run the default pypyr
    steps, on_success & on_failure sequence.

    If groups specified, will only run groups, without a success or failure
    sequence, unless you specifically set these also.

    Returns:
        None

    Raises:
        pypyr.errors.KeyNotInContextError: if ['pype'] or ['pype']['name']
                                           is missing.
        pypyr.errors.KeyInContextHasNoValueError: ['pype']['name'] exists but
                                                  is empty.
    """
    logger.debug("started")

    (pipeline_name,
     args,
     out,
     use_parent_context,
     pipe_arg,
     skip_parse,
     raise_error,
     loader,
     step_groups,
     success_group,
     failure_group
     ) = get_arguments(context)

    try:
        if use_parent_context:
            logger.info("pyping %s, using parent context.", pipeline_name)

            if args:
                logger.debug("writing args into parent context...")
                context.update(args)

            pipelinerunner.load_and_run_pipeline(
                pipeline_name=pipeline_name,
                pipeline_context_input=pipe_arg,
                context=context,
                parse_input=not skip_parse,
                loader=loader,
                groups=step_groups,
                success_group=success_group,
                failure_group=failure_group
            )
        else:
            logger.info("pyping %s, without parent context.", pipeline_name)

            if args:
                child_context = Context(args)
            else:
                child_context = Context()

            child_context.pipeline_name = pipeline_name
            child_context.working_dir = context.working_dir

            pipelinerunner.load_and_run_pipeline(
                pipeline_name=pipeline_name,
                pipeline_context_input=pipe_arg,
                context=child_context,
                parse_input=not skip_parse,
                loader=loader,
                groups=step_groups,
                success_group=success_group,
                failure_group=failure_group
            )

            if out:
                write_child_context_to_parent(out=out,
                                              parent_context=context,
                                              child_context=child_context)

        logger.info("pyped %s.", pipeline_name)
    except (ControlOfFlowInstruction, Stop):
        # Control-of-Flow/Stop are instructions to go somewhere
        # else, not errors per se.
        raise
    except Exception as ex_info:
        # yes, yes, don't catch Exception. Have to, though, in order to swallow
        # errs if !raise_error
        logger.error("Something went wrong pyping %s. %s: %s",
                     pipeline_name, type(ex_info).__name__, ex_info)

        if raise_error:
            logger.debug("Raising original exception to caller.")
            raise
        else:
            logger.debug(
                "raiseError is False. Swallowing error in %s.", pipeline_name)

    logger.debug("done")


def get_arguments(context):
    """Parse arguments for pype from context and assign default values.

    Args:
        context: pypyr.context.Context. context is mandatory.

    Returns:
        tuple (pipeline_name, #str
               args, #dict
               out, #str or dict or list
               use_parent_context, #bool
               pipe_arg, #str
               skip_parse, #bool
               raise_error #bool
               groups #list of str
               success_group #str
               failure_group #str
               )

    Raises:
       pypyr.errors.KeyNotInContextError: if ['pype']['name'] is missing.
       pypyr.errors.KeyInContextHasNoValueError: if ['pype']['name'] exists but
                                                 is None.
    """
    context.assert_key_has_value(key='pype', caller=__name__)
    pype = context.get_formatted('pype')

    try:
        pipeline_name = pype['name']

        if pipeline_name is None:
            raise KeyInContextHasNoValueError(
                "pypyr.steps.pype ['pype']['name'] exists but is empty.")
    except KeyError as err:
        raise KeyNotInContextError(
            "pypyr.steps.pype missing 'name' in the 'pype' context item. "
            "You need to specify the pipeline name to run another "
            "pipeline.") from err

    args = pype.get('args', None)

    if args is not None and not isinstance(args, dict):
        raise ContextError(
            "pypyr.steps.pype 'args' in the 'pype' context item "
            "must be a dict.")

    if args and 'useParentContext' not in pype:
        use_parent_context = False
    else:
        use_parent_context = pype.get('useParentContext', True)

    out = pype.get('out', None)
    if out and use_parent_context:
        raise ContextError(
            "pypyr.steps.pype pype.out is only relevant if useParentContext "
            "= False. If you're using the parent context, no need to have out "
            "args since their values will already be in context. If you're "
            "NOT using parent context and you've specified pype.args, just "
            "leave off the useParentContext key and it'll default to False "
            "under the hood, or set it to False yourself if you keep it in.")

    pipe_arg = pype.get('pipeArg', None)
    skip_parse = pype.get('skipParse', True)
    raise_error = pype.get('raiseError', True)
    loader = pype.get('loader', None)
    groups = pype.get('groups', None)
    if isinstance(groups, str):
        groups = [groups]

    success_group = pype.get('success', None)
    failure_group = pype.get('failure', None)

    return (
        pipeline_name,
        args,
        out,
        use_parent_context,
        pipe_arg,
        skip_parse,
        raise_error,
        loader,
        groups,
        success_group,
        failure_group
    )


def write_child_context_to_parent(out, parent_context, child_context):
    """Write out keys from child to parent context.

    Args:
        out. str or dict or list. Pass a string for a single
             key to grab from child context, a list of string for a list
             of keys to grab from child context, or a dict where you map
             'parent-key-name': 'child-key-name'.
        parent_context: parent Context. destination context.
        child_context: write from this context to the parent.
    """
    if isinstance(out, str):
        save_me = {out: out}
    elif isinstance(out, list):
        save_me = {k: k for k in out}
    elif isinstance(out, dict):
        save_me = out
    else:
        raise ContextError("pypyr.steps.pype pype.out should be a string, or "
                           f"a list or a dict. Instead, it's a {type(out)}")

    for parent_key, child_key in save_me.items():
        logger.debug(
            "setting parent context %s to value from child context %s",
            parent_key,
            child_key)

        parent_context[parent_key] = child_context.get_formatted(child_key)



"""Context parser that returns a dictionary from a local json file."""
from collections.abc import Mapping
import logging
import json

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)


def get_parsed_context(args):
    """Parse args as path to a json file and returns context as dictionary."""
    logger.debug("starting")
    if not args:
        raise AssertionError(
            "pipeline must be invoked with context arg set. For "
            "this json parser you're looking for something like:\n"
            "pypyr pipelinename ./myjsonfile.json")

    path = ' '.join(args)
    # open the json file on disk so that you can initialize the dictionary
    logger.debug("attempting to open file: %s", path)
    with open(path) as json_file:
        payload = json.load(json_file)

    if not isinstance(payload, Mapping):
        raise TypeError("json input should describe an object at the top "
                        "level. You should have something like\n"
                        "{\n\"key1\":\"value1\",\n\"key2\":\"value2\"\n}\n"
                        "at the json top-level, not an [array] or literal.")

    logger.debug("json file loaded into context. Count: %d", len(payload))
    logger.debug("done")
    return payload



"""pypyr pipeline yaml definition classes - domain specific language."""
import json
import logging
from ruamel.yaml.comments import CommentedMap, CommentedSeq
from ruamel.yaml.nodes import ScalarNode
from pypyr.errors import (Call,
                          ControlOfFlowInstruction,
                          get_error_name,
                          HandledError,
                          LoopMaxExhaustedError,
                          PipelineDefinitionError,
                          Stop)
from pypyr.cache.stepcache import step_cache
from pypyr.utils import expressions, poll

# use pypyr logger to ensure loglevel is set correctly
logger = logging.getLogger(__name__)

# ------------------------ custom yaml tags -----------------------------------


class SpecialTagDirective:
    """Base class for all special tag directives.

    Derive all special tag directives from this base class. Of particular
    interest is the get_value implementation.

    Derived classes MUST have the following attributes:
        - yaml_tag class variable. this is the yaml tag used to represent this
          class.
        - get_value() method override. this is what pypyr calls when applying
          formatting to this scalar.

    If your derived class wants to instantiate some instance properties, do so
    in __init__ rather than from_yaml. That way, self.value always contains the
    original, untouched scalar value.

    """

    def __init__(self, value):
        """Initialize class."""
        self.value = value

    def __bool__(self):
        """Truth testing - instance is falsy if no value."""
        return bool(self.value)

    def __str__(self):
        """Friendly string representation of instance."""
        return str(self.value)

    def __repr__(self):
        """Repr representation of instance."""
        return f'{self.__class__.__name__}({self.value!r})'

    def __eq__(self, other):
        """Equivalence over-ride."""
        return repr(self) == repr(other)

    @classmethod
    def to_yaml(cls, representer, node):
        """How to serialize this class back to yaml."""
        return representer.represent_scalar(cls.yaml_tag, node.value)

    @classmethod
    def from_yaml(cls, constructor, node):
        """Implement how to create the class from yaml representation."""
        return cls(node.value)

    def get_value(self, context=None):
        """Process the input value and return the output.

        Derived classes must provide an implementation for this method.

        This method is called by the various context get_formatted functions,
        where each special tag has its own processing rules about how to format
        values for that particular tag.
        """
        raise NotImplementedError(
            'Implement this to provide the processed value of your custom tag '
            'during formatting operations.')


class Jsonify(SpecialTagDirective):
    """Serialize context structure into a json string.

    Runs formatting substitution expressions on all formattable fields in the
    input structure.
    """

    yaml_tag = '!jsonify'

    def __init__(self, value, scalar=None):
        """Initialize class with special handling for TaggedScalar.

        Args:
            value (any): Value of yaml object. Likely CommentedMap or
                         CommentedSeq or if scalar the mapped Python type for
                         the scalar.
            scalar (TaggedScalar): If type scalar, the original constructed
                                   object. This is necessary for to_yaml
                                   serialization.
        """
        self.scalar = scalar
        self.value = value

    def __repr__(self):
        """Handle TaggedScalar specially.

        This is because original node necessary to reconstruct a TaggedScalar
        and .value gives the underlying simple type value instead.
        """
        if self.scalar:
            return (
                f'{self.__class__.__name__}({self.value!r}, {self.scalar!r})')
        else:
            return f'{self.__class__.__name__}({self.value!r})'

    @classmethod
    def from_yaml(cls, constructor, node):
        """Create the class from yaml representation."""
        for data in constructor.construct_undefined(node):
            # returns generator as a means to update the
            # returned iterator for recursive yaml refs where the node is still
            # under construction.
            pass

        if isinstance(node, ScalarNode):
            if node.style:
                # style having a value means it is quoted, i.e always a str
                # therefore don't need to go and hunt down the type
                return cls(data.value, data)

            # constructed_undefined creates all scalar as TaggedScalar, which
            # is always str. Use resolver to construct object to get it to
            # parse to the appropriate Python literal simple type.
            tag = constructor.resolver.resolve(ScalarNode,
                                               node.value,
                                               (True, False))

            scalar_node = ScalarNode(tag,
                                     node.value,
                                     start_mark=node.start_mark,
                                     end_mark=node.end_mark,
                                     style=node.style,
                                     comment=node.comment,
                                     anchor=node.anchor)

            constructed_scalar_node = constructor.construct_object(scalar_node)
            return cls(constructed_scalar_node, data)

        return cls(data)

    @classmethod
    def to_yaml(cls, representer, node):
        """Serialize this class back to yaml."""
        if isinstance(node.value, CommentedMap):
            return representer.represent_mapping(cls.yaml_tag, node.value)
        elif isinstance(node.value, CommentedSeq):
            return representer.represent_sequence(cls.yaml_tag, node.value)
        else:
            return representer.represent_tagged_scalar(node.scalar)

    def get_value(self, context):
        """Serialize self contents to json."""
        # dumps writes null if self.value is None
        return json.dumps(context.get_formatted_value(self.value))


class PyString(SpecialTagDirective):
    """py strings allow you to execute python expressions dynamically.

    A py string is defined by custom yaml tag like this:
    !py <<<your expression here>>>

    For example:
        !py key == 3

    Will return True if context['key'] is 3.

    This provides dynamic python eval of an input expression. The return is
    whatever the result of the expression is.

    Use with caution: since input_string executes any arbitrary code object
    the potential for damage is great.

    The eval uses the current context object as the namespace. This means
    if you have context['mykey'], in the input_string expression you can
    use the key directly as a variable like this: "mykey == 'mykeyvalue'".

    Unlike formatting expressions, key names do NOT go in {curlies}.

    Both __builtins__ and context are available to the eval expression.

    """

    yaml_tag = '!py'

    def get_value(self, context):
        """Run python eval on the input string."""
        if self.value:
            return expressions.eval_string(self.value, context)
        else:
            # Empty input raises cryptic EOF syntax err, this more human
            # friendly
            raise ValueError('!py string expression is empty. It must be a '
                             'valid python expression instead.')


class SicString(SpecialTagDirective):
    """Sic strings do not have any processing applied to them.

    If a string is NOT to have {substitutions} run on it, it's sic erat
    scriptum, i.e literal.

    A sic string looks like this:
    input_string=!sic <<your string literal here>>

    For example:
        !sic piping {key} the valleys wild

    Will return "piping {key} the valleys wild" without attempting to
    substitute {key} from context.

    """

    yaml_tag = '!sic'

    def get_value(self, context=None):
        """Simply return the string as is, the whole point of a sic string."""
        return self.value

# ------------------------ END custom yaml tags -------------------------------


class Step:
    """A step, as interpreted by the pypyr pipeline definition yaml.

    Encapsulates useful methods for the execution of a step, and also
    maintains state necessary to run a step. Given the need to maintain state,
    this is in a class, rather than purely functional code.

    Dynamically loads the module that the step points at, interprets the step
    decorators that modify step execution, and runs the actual step with
    appropriate error handling.

    External class consumers should probably use the run_step method. run_step
    serves as the blackbox entrypoint for this class' other methods.

    Attributes:
        name: (string) this is the step-name. equivalent to the module name of
              of the step. this module is the one dynamically loaded to
              the module attribute.
        module: (importlib module) the dynamically loaded module that the
                step will execute. this module will have the run_step
                function that implements the actual step execution.
        foreach_items: (list) defaults None. Execute step once for each item in
                    list, using iterator i.
        in_parameters: (dict) defaults None. The in step decorator - i.e dict
                       to add to context before step execution.
        run_me: (bool) defaults True. step runs if this is true.
        skip_me: (bool) defaults False. step does not run if this is true.
        steps_runner: pypyr.stepsrunner.StepsRunner Step Runner instance
                      running this step.
        swallow_me: (bool) defaults False. swallow any errors during step run
                    and continue processing if true.
        while_decorator: (WhileDecorator) defaults None. execute step in while
                         loop.

    """

    def __init__(self, step, steps_runner):
        """Initialize the class. No duh, huh?.

        You can happily expect the initializer to initialize all
        member attributes.

        Args:
            step: a string or a dict. This is the actual step as it exists in
                  the pipeline yaml - which is to say it can just be a string
                  for a simple step, or a dict for a complex step.
            steps_runner: the StepsRunner instance running this Step.

        """
        logger.debug("starting")

        self.steps_runner = steps_runner

        # defaults for decorators
        self.description = None
        self.foreach_items = None
        self.in_parameters = None
        self.retry_decorator = None
        self.line_no = None
        self.line_col = None
        self.run_me = True
        self.skip_me = False
        self.swallow_me = False
        self.name = None
        self.while_decorator = None
        self.on_error = None

        try:
            if isinstance(step, dict):
                self._init_from_dict(step)
            else:
                # of course, it might not be a string. in line with duck
                # typing, beg forgiveness later. as long as it loads
                # the module, happy days.
                logger.debug("%s is a simple string.", step)
                self.name = step

            self.run_step_function = step_cache.get_step(self.name)
        except Exception:
            # Exceptions could also happened on the step init phase
            # (ModuleNotFound, KeyError, etc..),
            # put exception handler here because of that.
            # also handle case with missing step name
            name = f" {self.name}" if self.name else ""

            if self.line_no:
                logger.error(
                    "Error at pipeline step%s yaml line: "
                    "%d, col: %d",
                    name, self.line_no, self.line_col
                )
            else:
                logger.error(
                    "Error at pipeline step%s", name
                )
            raise

        logger.debug("done")

    def _init_from_dict(self, step):
        """Initialize the class from a dict for a complex step.

        Args:
            step: (CommentedMap/dict) This is the actual step as it
            exists in the pipeline yaml.
        """
        if hasattr(step, 'lc'):
            # line_no: optional. Has value only when the yaml
            # round trip parser is in use.
            self.line_no = step.lc.line + 1
            # line_col: optional. Has value only when the yaml
            # round trip parser is in use.
            self.line_col = step.lc.col + 1

        self.name = step.get('name', None)

        if not self.name:
            raise PipelineDefinitionError('step must have a name.')

        logger.debug("%s is complex.", self.name)

        self.in_parameters = step.get('in', None)

        # description: optional. Write to stdout if exists and flagged.
        self.description = step.get('description', None)

        # foreach: optional value. None by default.
        self.foreach_items = step.get('foreach', None)
        if self.foreach_items:
            # current item in loops
            self.for_counter = None

        # retry: optional, defaults none.
        retry_definition = step.get('retry', None)
        if retry_definition:
            self.retry_decorator = RetryDecorator(retry_definition)

        # run: optional value, true by default. Allow substitution.
        self.run_me = step.get('run', True)

        # skip: optional value, false by default. Allow substitution.
        self.skip_me = step.get('skip', False)

        # swallow: optional, defaults false. Allow substitution.
        self.swallow_me = step.get('swallow', False)

        # on_error: optional, defaults none. Allow substitution.
        self.on_error = step.get('onError', None)

        # while: optional, defaults none.
        while_definition = step.get('while', None)
        if while_definition:
            self.while_decorator = WhileDecorator(while_definition)

        logger.debug("step name: %s", self.name)

    def save_error(self, context, exception, swallowed):
        """Append step's exception information to the context.

        Append the[on_error] dictionary to the context. This will append to
        existing `runErrors` values if `runErrors` are already in there.

        Args:
            context: (pypyr.context.Context) The pypyr context. This arg will
                     mutate - after method execution will contain the new
                     updated context.
            exception: (Exception) The error detected during step execution.
            swallowed: (bool) Whether exception was swallowed or not.
        """
        failure = {
            'name': get_error_name(exception),
            'description': str(exception),
            'customError': context.get_formatted_value(
                self.on_error
            ) if self.on_error else {},
            'line': self.line_no,
            'col': self.line_col,
            'step': self.name,
            'exception': exception,
            'swallowed': swallowed,
        }

        run_errors = context.setdefault('runErrors', [])

        run_errors.append(failure)

    def foreach_loop(self, context):
        """Run step once for each item in foreach_items.

        On each iteration, the invoked step can use context['i'] to get the
        current iterator value.

        Args:
            context: (pypyr.context.Context) The pypyr context. This arg will
                     mutate.
        """
        logger.debug("starting")

        # Loop decorators only evaluated once, not for every step repeat
        # execution.
        foreach = context.get_formatted_value(self.foreach_items)

        foreach_length = len(foreach)

        logger.info("foreach decorator will loop %s times.", foreach_length)

        for i in foreach:
            logger.info("foreach: running step %s", i)
            # the iterator must be available to the step when it executes
            context['i'] = i
            self.for_counter = i

            # conditional operators apply to each iteration, so might be an
            # iteration run, skips or swallows.
            self.run_conditional_decorators(context)
            logger.debug("foreach: done step %s", i)

        logger.debug("foreach decorator looped %s times.", foreach_length)
        logger.debug("done")

    def invoke_step(self, context):
        """Invoke 'run_step' in the dynamically loaded step module.

        Don't invoke this from outside the Step class. Use
        pypyr.dsl.Step.run_step instead.
        invoke_step just does the bare module step invocation, it does not
        evaluate any of the decorator logic surrounding the step. So unless
        you really know what you're doing, use run_step if you intend on
        executing the step the same way pypyr does.

        Args:
            context: (pypyr.context.Context) The pypyr context. This arg will
                     mutate.
        """
        logger.debug("starting")

        logger.debug("running step %s", self.name)

        try:
            self.run_step_function(context)
        except Call as call:
            logger.debug("call: calling %s", call.groups)
            try:
                self.steps_runner.run_step_groups(
                    groups=call.groups,
                    success_group=call.success_group,
                    failure_group=call.failure_group)
            except Exception as ex_info:
                # don't want to log error twice - would've been logged already
                # in called step-group.
                raise HandledError from ex_info
            finally:
                self.reset_context_counters(context, call)

            # py 3.9 issue with coveragepy doesn't show this as covered. it is.
            logger.debug("call: done calling %s",
                         call.groups)  # pragma: no cover

        logger.debug("step %s done", self.name)

    def reset_context_counters(self, context, call):
        """Set loop counters in context to current counters on self.

        Set the following context properties to current:
        - whileCounter
        - i
        - retryCounter

        Also resets pypyr.steps.call config. In context 'call' MUST exist,
        because this method is only meant to be called stemming from a Call
        control-of-flow instruction.

        This method exists because in nested pypyr.steps.call invocations
        the child overwrites the parent's context config and similarly nested
        loop counters get overwritten. So when the called child group is done,
        use this method will reset context to continue with the parent.

        This is necessary for nested calls in loops. A nested call step's
        context config (under key 'call') will override the parent's, so when
        the nested call finishes, in order to continue the parent call you need
        to reset the config under key 'call' to what it was before the child
        got its hands on it.

        Args:
        context: (pypyr.context.Context) The pypyr context. This arg will
                 mutate.
        call: pypyr.errors.Call The control-of-flow call instruction.

        """
        if self.while_decorator:
            while_counter = self.while_decorator.while_counter
            if context['whileCounter'] != while_counter:
                context['whileCounter'] = while_counter

        if self.foreach_items:
            # an individual item could be None, so no bool check on
            # counter itself, use foreach_items instead.
            if context['i'] != self.for_counter:
                context['i'] = self.for_counter

        if self.retry_decorator:
            retry_counter = self.retry_decorator.retry_counter
            if context['retryCounter'] != retry_counter:
                context['retryCounter'] = retry_counter

        # if call does not have a value something is seriously wrong.
        assert call.original_config[1]

        # this might be a nested call instruction, so a child call
        # might have overwritten the 'call' context item.
        if context.get(call.original_config[0]) is not call.original_config[1]:
            context[call.original_config[0]] = call.original_config[1]

    def run_conditional_decorators(self, context):
        """Evaluate the step decorators to decide whether to run step or not.

        Use pypyr.dsl.Step.run_step if you intend on executing the step the
        same way pypyr does.

        Args:
            context: (pypyr.context.Context) The pypyr context. This arg will
                     mutate.
        """
        logger.debug("starting")

        # The decorator attributes might contain formatting expressions that
        # change whether they evaluate True or False, thus apply formatting at
        # last possible instant.
        run_me = context.get_formatted_as_type(self.run_me, out_type=bool)
        skip_me = context.get_formatted_as_type(self.skip_me, out_type=bool)
        swallow_me = context.get_formatted_as_type(self.swallow_me,
                                                   out_type=bool)

        if run_me:
            if not skip_me:
                try:
                    if self.retry_decorator:
                        self.retry_decorator.retry_loop(context,
                                                        self.invoke_step)
                    else:
                        self.invoke_step(context=context)
                except (ControlOfFlowInstruction, Stop):
                    # Control-of-Flow/Stop are instructions to go somewhere
                    # else, not errors per se.
                    raise
                except Exception as exc_info:
                    if isinstance(exc_info, HandledError):
                        exc_info = exc_info.__cause__
                    else:
                        # prevent already logged err logging twice.
                        self.save_error(
                            context=context,
                            exception=exc_info,
                            swallowed=swallow_me
                        )
                    if swallow_me:
                        logger.error(
                            "%s Ignoring error because swallow "
                            "is True for this step.\n"
                            "%s: %s",
                            self.name, get_error_name(exc_info), exc_info
                        )
                    else:
                        if self.line_no:
                            logger.error(
                                "Error while running step %s "
                                "at pipeline yaml line: %d, col: %d",
                                self.name, self.line_no, self.line_col
                            )
                        else:
                            logger.error(
                                "Error while running step %s", self.name
                            )
                        raise exc_info
            else:
                logger.info(
                    "%s not running because skip is True.", self.name)
        else:
            logger.info("%s not running because run is False.", self.name)

        logger.debug("done")

    def run_foreach_or_conditional(self, context):
        """Run the foreach sequence or the conditional evaluation.

        Args:
            context: (pypyr.context.Context) The pypyr context. This arg will
                     mutate.
        """
        logger.debug("starting")
        # friendly reminder [] list obj (i.e empty) evals False
        if self.foreach_items:
            self.foreach_loop(context)
        else:
            # since no looping required, don't pollute output with looping info
            self.run_conditional_decorators(context)

        logger.debug("done")

    def run_step(self, context):
        """Run a single pipeline step.

        Args:
            context: (pypyr.context.Context) The pypyr context. This arg will
                     mutate.
        """
        logger.debug("starting")

        # give user helpful output if step will actually run or not.
        if self.description:
            description = context.get_formatted_value(self.description)
            run_me = context.get_formatted_as_type(self.run_me, out_type=bool)
            skip_me = context.get_formatted_as_type(self.skip_me,
                                                    out_type=bool)

            if run_me and not skip_me:
                logger.notify(description)
            else:
                logger.notify("(skipping): %s", description)

        # the in params should be added to context before step execution.
        self.set_step_input_context(context)

        if self.while_decorator:
            self.while_decorator.while_loop(context,
                                            self.run_foreach_or_conditional)
        else:
            self.run_foreach_or_conditional(context)

        # the in params should be removed from context after step execution.
        self.unset_step_input_context(context)

        logger.debug("done")

    def set_step_input_context(self, context):
        """Append step's 'in' parameters to context, if they exist.

        Append the [in] dictionary to the context. This will overwrite
        existing values if the same keys are already in there. I.e if
        in_parameters has {'eggs': 'boiled'} and key 'eggs' already
        exists in context, context['eggs'] hereafter will be 'boiled'.

        Args:
            context: (pypyr.context.Context) The pypyr context. This arg will
                     mutate - after method execution will contain the new
                     updated context.
        """
        logger.debug("starting")
        if self.in_parameters is not None:
            parameter_count = len(self.in_parameters)
            if parameter_count > 0:
                logger.debug(
                    "Updating context with %s 'in' parameters.",
                    parameter_count,
                )
                context.update(self.in_parameters)

        logger.debug("done")

    def unset_step_input_context(self, context):
        """Remove step's 'in' parameters from context, if they exist.

        Remove the [in] dictionary from the context.

        Args:
            context: (pypyr.context.Context) The pypyr context. This arg will
                     mutate - after method execution will contain the new
                     updated context with the [in] args removed.
        """
        logger.debug("starting")

        if self.in_parameters is not None:
            # len is O(1), so not all that unnecessarily duplicated cpu time
            parameter_count = len(self.in_parameters)
            if parameter_count > 0:
                logger.debug(
                    "Removing %s 'in' parameters from context.",
                    parameter_count,
                )
                for key in self.in_parameters:
                    # slightly unorthodox pop returning None means you don't
                    # get a KeyError if key doesn't exist. The step might have
                    # removed the key from context.
                    context.pop(key, None)

        logger.debug("done")


class RetryDecorator:
    """Retry decorator, as interpreted by the pypyr pipeline definition yaml.

    Encapsulate the methods that run a step in a retry loop, and also maintains
    state necessary to run the loop. Given the need to maintain state, this is
    in a class, rather than purely functional code.

    In a normal world, Step invokes RetryDecorator. If you run it directly,
    you're responsible for the context and surrounding control-of-flow.

    External class consumers should probably use the retry_loop method.
    retry_loop serves as the blackbox entrypoint for this class' other methods.

    Attributes:
        max: (int) default None. Maximum loop iterations. None is infinite.
        sleep: (float) defaults 0. Sleep in seconds between iterations.
        stop_on: (list) default None. Always stop retry on these error
                 types. None means retry on all errors.
        retry_on: (list) default None. Only retry on these error types. All
                other error types will stop retry loop. None means retry all
                errors.

    """

    def __init__(self, retry_definition):
        """Initialize the class. No duh, huh.

        You can happily expect the initializer to initialize all
        member attributes.

        Args:
            retry_definition: dict. This is the actual retry definition as it
                              exists in the pipeline yaml.

        """
        logger.debug("starting")

        if isinstance(retry_definition, dict):
            # max: optional. defaults None.
            self.max = retry_definition.get('max', None)

            # sleep: optional. defaults 0.
            self.sleep = retry_definition.get('sleep', 0)

            # stopOn: optional. defaults None.
            self.stop_on = retry_definition.get('stopOn', None)

            # retryOn: optional. defaults None.
            self.retry_on = retry_definition.get('retryOn', None)
        else:
            # if it isn't a dict, pipeline configuration is wrong.
            logger.error("retry decorator definition incorrect.")
            raise PipelineDefinitionError("retry decorator must be a dict "
                                          "(i.e a map) type.")

        self.retry_counter = None

        logger.debug("done")

    def exec_iteration(self, counter, context, step_method):
        """Run a single retry iteration.

        This method abides by the signature invoked by poll.while_until_true,
        which is to say (counter, *args, **kwargs). In a normal execution
        chain, this method's args passed by self.retry_loop where context
        and step_method set. while_until_true injects counter as a 1st arg.

        Args:
            counter. int. loop counter, which number of iteration is this.
            context: (pypyr.context.Context) The pypyr context. This arg will
                     mutate - after method execution will contain the new
                     updated context.
            step_method: (method/function) This is the method/function that
                         will execute on every loop iteration. Signature is:
                         function(context)

         Returns:
            bool. True if step execution completed without error.
                  False if error occured during step execution.

        """
        logger.debug("starting")
        context['retryCounter'] = counter
        self.retry_counter = counter

        logger.info("retry: running step with counter %s", counter)
        try:
            step_method(context)
            result = True
        except (ControlOfFlowInstruction, Stop):
            # Control-of-Flow/Stop are instructions to go somewhere
            # else, not errors per se.
            raise
        except Exception as ex_info:
            if self.max:
                if counter == self.max:
                    logger.debug("retry: max %s retries exhausted. "
                                 "raising error.", counter)
                    # arguably shouldn't be using errs for control of flow.
                    # but would lose the err info if not, so lesser of 2 evils.
                    raise

            if isinstance(ex_info, HandledError):
                ex_info = ex_info.__cause__

            if self.stop_on or self.retry_on:
                error_name = get_error_name(ex_info)
                if self.stop_on:
                    formatted_stop_list = context.get_formatted_value(
                        self.stop_on)
                    if error_name in formatted_stop_list:
                        logger.error("%s in stopOn. Raising error "
                                     "and exiting retry.", error_name)
                        raise
                    else:
                        logger.debug("%s not in stopOn. Continue.", error_name)

                if self.retry_on:
                    formatted_retry_list = context.get_formatted_value(
                        self.retry_on)
                    if error_name not in formatted_retry_list:
                        logger.error("%s not in retryOn. Raising "
                                     "error and exiting retry.", error_name)
                        raise
                    else:
                        logger.debug("%s in retryOn. Retry again.", error_name)

            result = False
            logger.error("retry: ignoring error because retryCounter < max.\n"
                         "%s: %s", type(ex_info).__name__, ex_info)

        logger.debug("retry: done step with counter %s", counter)

        logger.debug("done")
        return result

    def retry_loop(self, context, step_method):
        """Run step inside a retry loop.

        Args:
            context: (pypyr.context.Context) The pypyr context. This arg will
                     mutate - after method execution will contain the new
                     updated context.
            step_method: (method/function) This is the method/function that
                         will execute on every loop iteration. Signature is:
                         function(context)

        """
        logger.debug("starting")

        context['retryCounter'] = 0
        self.retry_counter = 0

        sleep = context.get_formatted_as_type(self.sleep, out_type=float)
        if self.max:
            max = context.get_formatted_as_type(self.max, out_type=int)

            logger.info("retry decorator will try %d times at %ss "
                        "intervals.", max, sleep)
        else:
            max = None
            logger.info("retry decorator will try indefinitely at %ss "
                        "intervals.", sleep)

        # this will never be false. because on counter == max,
        # exec_iteration raises an exception, breaking out of the loop.
        # pragma because cov doesn't know the implied else is impossible.
        # unit test cov is 100%, though.
        if poll.while_until_true(interval=sleep,
                                 max_attempts=max)(
                self.exec_iteration)(context=context,
                                     step_method=step_method
                                     ):  # pragma: no cover
            logger.debug("retry loop complete, reporting success.")

        logger.debug("retry loop done")

        logger.debug("done")


class WhileDecorator:
    """While Decorator, as interpreted by the pypyr pipeline definition yaml.

    Encapsulate the methods that run a step in a while loop, and also maintains
    state necessary to run the loop. Given the need to maintain state, this is
    in a class, rather than purely functional code.

    In a normal world, Step invokes WhileDecorator. If you run it directly,
    you're responsible for the context and surrounding control-of-flow.

    External class consumers should probably use the while_loop method.
    while_loop serves as the blackbox entrypoint for this class' other methods.

    Attributes:
        error_on_max: (bool) defaults False. Raise error if max reached.
        max: (int) default None. Maximum loop iterations. None is infinite.
        sleep: (float) defaults 0. Sleep in seconds between iterations.
        stop:(bool) defaults None. Exit loop when stop is True.

    """

    def __init__(self, while_definition):
        """Initialize the class. No duh, huh.

        You can happily expect the initializer to initialize all
        member attributes.

        Args:
            while_definition: dict. This is the actual while definition as it
                              exists in the pipeline yaml.

        """
        logger.debug("starting")

        if isinstance(while_definition, dict):
            # errorOnMax: optional. defaults False
            self.error_on_max = while_definition.get('errorOnMax', False)

            # max: optional. defaults None.
            self.max = while_definition.get('max', None)

            # sleep: optional. defaults 0.
            self.sleep = while_definition.get('sleep', 0)

            # stop: optional. defaults None.
            self.stop = while_definition.get('stop', None)

            if self.stop is None and self.max is None:
                logger.error("while decorator missing both max and stop.")
                raise PipelineDefinitionError("the while decorator must have "
                                              "either max or stop, or both. "
                                              "But not neither. Note that "
                                              "setting stop: False with no "
                                              "max is an infinite loop. If "
                                              "an infinite loop is really "
                                              "what you want, set stop: False")
        else:
            # if it isn't a dict, pipeline configuration is wrong.
            logger.error("while decorator definition incorrect.")
            raise PipelineDefinitionError("while decorator must be a dict "
                                          "(i.e a map) type.")

        self.while_counter = None

        logger.debug("done")

    def exec_iteration(self, counter, context, step_method):
        """Run a single loop iteration.

        This method abides by the signature invoked by poll.while_until_true,
        which is to say (counter, *args, **kwargs). In a normal execution
        chain, this method's args passed by self.while_loop where context
        and step_method set. while_until_true injects counter as a 1st arg.

        Args:
            counter. int. loop counter, which number of iteration is this.
            context: (pypyr.context.Context) The pypyr context. This arg will
                     mutate - after method execution will contain the new
                     updated context.
            step_method: (method/function) This is the method/function that
                         will execute on every loop iteration. Signature is:
                         function(context)

         Returns:
            bool. True if self.stop evaluates to True after step execution,
                  False otherwise.

        """
        logger.debug("starting")
        context['whileCounter'] = counter
        self.while_counter = counter

        logger.info("while: running step with counter %s", counter)
        step_method(context)
        logger.debug("while: done step %s", counter)

        result = False
        # if no stop, just iterating to max)
        if self.stop:
            # dynamically evaluate stop after step execution, since the step
            # might have changed True/False status for stop.
            result = context.get_formatted_as_type(self.stop, out_type=bool)

        logger.debug("done")
        return result

    def while_loop(self, context, step_method):
        """Run step inside a while loop.

        Args:
            context: (pypyr.context.Context) The pypyr context. This arg will
                     mutate - after method execution will contain the new
                     updated context.
            step_method: (method/function) This is the method/function that
                         will execute on every loop iteration. Signature is:
                         function(context)

        """
        logger.debug("starting")

        context['whileCounter'] = 0
        self.while_counter = 0

        if self.stop is None and self.max is None:
            # the ctor already does this check, but guess theoretically
            # consumer could have messed with the props since ctor
            logger.error("while decorator missing both max and stop.")
            raise PipelineDefinitionError("the while decorator must have "
                                          "either max or stop, or both. "
                                          "But not neither.")

        error_on_max = context.get_formatted_as_type(
            self.error_on_max, out_type=bool)
        sleep = context.get_formatted_as_type(self.sleep, out_type=float)
        if self.max is None:
            max = None
            logger.info("while decorator will loop until %s "
                        "evaluates to True at %ss intervals.",
                        self.stop, sleep)
        else:
            max = context.get_formatted_as_type(self.max, out_type=int)

            if max < 1:
                logger.info(
                    "max %s is %s. while only runs when max > 0.",
                    self.max, max)
                logger.debug("done")
                return

            if self.stop is None:
                logger.info("while decorator will loop %s times at "
                            "%ss intervals.", max, sleep)
            else:
                logger.info("while decorator will loop %s times, or "
                            "until %s evaluates to True at "
                            "%ss intervals.", max, self.stop, sleep)

        if not poll.while_until_true(interval=sleep,
                                     max_attempts=max)(
                self.exec_iteration)(context=context,
                                     step_method=step_method):
            # False means loop exhausted and stop never eval-ed True.
            if error_on_max:
                logger.error("exhausted %s iterations of while loop, "
                             "and errorOnMax is True.", max)
                if self.stop and max:
                    raise LoopMaxExhaustedError("while loop reached "
                                                f"{max} and {self.stop} "
                                                "never evaluated to True.")
                else:
                    raise LoopMaxExhaustedError(f"while loop reached {max}.")
            else:
                if self.stop and max:
                    logger.info(
                        "while decorator looped %s times, "
                        "and %s never evaluated to True.", max, self.stop)

            logger.debug("while loop done")
        else:
            logger.info("while loop done, stop condition %s "
                        "evaluated True.", self.stop)

        logger.debug("done")



"""pypyr context class. Dictionary ahoy."""
from collections import namedtuple
from collections.abc import Mapping, Set
from pypyr.dsl import SpecialTagDirective
from pypyr.errors import KeyInContextHasNoValueError, KeyNotInContextError
from pypyr.formatting import RecursiveFormatter
from pypyr.utils import asserts, expressions, types

ContextItemInfo = namedtuple('ContextItemInfo',
                             ['key',
                              'key_in_context',
                              'expected_type',
                              'is_expected_type',
                              'has_value'])


class Context(dict):
    """The pypyr context.

    This is a mutable dict that maintains state during the entire life-span of
    a pipeline.

    This class only adds functionality on top of dictionary, it should not
    override anything in dict.

    Attributes:
        pipeline_name (str): name of pipeline that is currently running
        working_dir (path-like): working directory path. Either CWD or
                                 initialized from the cli --dir arg.

    """

    # I *think* instantiating formatter at class level is fine - far as I can
    # see the class methods are functional, not dependant on class state (also,
    # no __init__).
    # https://github.com/python/cpython/blob/master/Lib/string.py
    formatter = RecursiveFormatter(special_types=SpecialTagDirective)

    def __missing__(self, key):
        """Throw KeyNotInContextError rather than KeyError.

        Python explicitly clears this over-ride for dict inheritance.
        https://docs.python.org/3/library/stdtypes.html#dict
        """
        raise KeyNotInContextError(f"{key} not found in the pypyr context.")

    def assert_child_key_has_value(self, parent, child, caller):
        """Assert that context contains key that has child which has a value.

        Args:
            parent: parent key
            child: validate this sub-key of parent exists AND isn't None.
            caller: string. calling function name - this used to construct
                    error messages

        Raises:
            KeyNotInContextError: Key doesn't exist
            KeyInContextHasNoValueError: context[key] is None
            AssertionError: if key is None

        """
        asserts.assert_key_has_value(self,
                                     parent,
                                     caller)
        asserts.assert_key_has_value(self[parent],
                                     child,
                                     caller,
                                     parent)

    def assert_key_exists(self, key, caller):
        """Assert that context contains key.

        Args:
            key: validates that this key exists in context
            caller: string. calling function or module name - this used to
                    construct error messages

        Raises:
            KeyNotInContextError: When key doesn't exist in context.

        """
        asserts.assert_key_exists(self, key, caller)

    def assert_key_has_value(self, key, caller):
        """Assert that context contains key which also has a value.

        Args:
            key: validate this key exists in context AND has a value that isn't
                 None.
            caller: string. calling function name - this used to construct
                    error messages

        Raises:
            KeyNotInContextError: Key doesn't exist
            KeyInContextHasNoValueError: context[key] is None
            AssertionError: if key is None

        """
        asserts.assert_key_has_value(self, key, caller)

    def assert_key_type_value(self,
                              context_item,
                              caller,
                              extra_error_text=''):
        """Assert that keys exist of right type and has a value.

        Args:
             context_item: ContextItemInfo tuple
             caller: string. calling function name - this used to construct
                     error messages
             extra_error_text: append to end of error message.

        Raises:
            AssertionError: if context_item None.
            KeyNotInContextError: Key doesn't exist
            KeyInContextHasNoValueError: context[key] is None or the wrong
                                         type.

        """
        assert context_item, ("context_item parameter must be specified.")

        if extra_error_text is None or extra_error_text == '':
            append_error_text = ''
        else:
            append_error_text = f' {extra_error_text}'

        if not context_item.key_in_context:
            raise KeyNotInContextError(f'{caller} couldn\'t find '
                                       f'{context_item.key} in context.'
                                       f'{append_error_text}')

        if not context_item.has_value:
            raise KeyInContextHasNoValueError(
                f'{caller} found {context_item.key} in '
                f'context but it doesn\'t have a value.'
                f'{append_error_text}')

        if not context_item.is_expected_type:
            raise KeyInContextHasNoValueError(
                f'{caller} found {context_item.key} in context, but it\'s '
                f'not a {context_item.expected_type}.'
                f'{append_error_text}')

    def assert_keys_exist(self, caller, *keys):
        """Assert that context contains keys.

        Args:
            keys: validates that these keys exists in context
            caller: string. calling function or module name - this used to
                    construct error messages

        Raises:
            KeyNotInContextError: When key doesn't exist in context.

        """
        assert keys, ("*keys parameter must be specified.")
        for key in keys:
            self.assert_key_exists(key, caller)

    def assert_keys_have_values(self, caller, *keys):
        """Check that keys list are all in context and all have values.

        Args:
            *keys: Will check each of these keys in context
            caller: string. Calling function name - just used for informational
                    messages

        Raises:
            KeyNotInContextError: Key doesn't exist
            KeyInContextHasNoValueError: context[key] is None
            AssertionError: if *keys is None

        """
        for key in keys:
            self.assert_key_has_value(key, caller)

    def assert_keys_type_value(self,
                               caller,
                               extra_error_text,
                               *context_items):
        """Assert that keys exist of right type and has a value.

        Args:
             caller: string. calling function name - this used to construct
                     error messages
             extra_error_text: append to end of error message. This can happily
                               be None or ''.
            *context_items: ContextItemInfo tuples

        Raises:
            AssertionError: if context_items None.
            KeyNotInContextError: Key doesn't exist
            KeyInContextHasNoValueError: context[key] is None or the wrong
                                         type.

        """
        assert context_items, ("context_items parameter must be specified.")

        for context_item in context_items:
            self.assert_key_type_value(context_item, caller, extra_error_text)

    def get_eval_string(self, input_string):
        """Dynamically evaluates the input_string expression.

        This provides dynamic python eval of an input expression. The return is
        whatever the result of the expression is.

        Use with caution: since input_string executes any arbitrary code object
        the potential for damage is great.

        The eval uses the current context object as the namespace. This means
        if you have context['mykey'], in the input_string expression you can
        use the key directly as a variable like this: "mykey == 'mykeyvalue'".

        Both __builtins__ and context are available to the eval expression.

        Args:
            input_string: expression to evaluate.

        Returns:
            Whatever object results from the string expression valuation.

        """
        return expressions.eval_string(input_string, self)

    def get_formatted(self, key):
        """Return formatted value for context[key].

        This is a convenience method that calls the same thing as
        get_formatted_value() under the hood, passing to it the value it
        retrieves from context at the input key.

        If context[key]'s value is a type string, will just format and return
        the string. Strings can contain recursive formatting expressions.

        If context[key]'s value is a special type, like a py string or sic
        string, will run the formatting implemented by the custom tag
        representer.

        If context[key] is not a string, specifically an iterable type like a
        dict, list, tuple, set, it will use get_formatted_value under the
        covers to loop through and handle the entire structure contained in
        context[key].

        If context[key]='Piping {key1} the {key2} wild'
        And context={'key1': 'down', 'key2': 'valleys', 'key3': 'value3'}

        Then this will return string: "Piping down the valleys wild"

        Choosing between get_formatted() and get_formatted_value():
        - get_formatted() gets a context[key] value with formatting applied.
        - get_formatted_value() is for any arbitrary object.

        Args:
            key: dictionary key to retrieve.

        Returns:
            Whatever object results from the formatting expression(s) at the
            input key's value.

        Raises:
            KeyNotInContextError: context[key] value contains {somekey} where
                                  somekey does not exist in context dictionary.

        """
        val = self[key]

        try:
            # any sort of complex type will work with recursive formatter.
            return self.formatter.vformat(val, None, self)
        except KeyNotInContextError as err:
            # less cryptic error for end-user friendliness
            raise KeyNotInContextError(
                f'Unable to format \'{val}\' at context[\'{key}\'], '
                f'because {err}'
            ) from err

    def get_formatted_iterable(self, obj, memo=None):
        """Use get_formatted_value(input_value) instead. Deprecated."""
        from warnings import warn
        warn(
            ("Use get_formatted_value(input_value) instead of "
             "get_formatted_iterable"),
            DeprecationWarning)
        return self.formatter.vformat(obj, None, self)

    def get_formatted_string(self, input_string):
        """Use get_formatted_value(input_value) instead. Deprecated."""
        from warnings import warn
        warn(
            ("Use get_formatted_value(input_value) instead of "
             "get_formatted_string"),
            DeprecationWarning)
        if isinstance(input_string, str):
            try:
                return self.formatter.vformat(input_string, None, self)
            except KeyNotInContextError as err:
                # Wrapping the KeyError into a less cryptic error for end-user
                # friendliness
                raise KeyNotInContextError(
                    f'Unable to format \'{input_string}\' because {err}'
                ) from err
        elif isinstance(input_string, SpecialTagDirective):
            return input_string.get_value(self)
        else:
            raise TypeError(f"can only format on strings. {input_string} is a "
                            f"{type(input_string)} instead.")

    def get_formatted_as_type(self, value, default=None, out_type=str):
        """Return formatted value for input value, returns as out_type.

        Caveat emptor: if out_type is bool and value a string,
        return will be True if str is 'True', 'TRUE', '1' or '1.0'. It will be
        False for all other cases.

        Args:
            value: the value to format
            default: if value is None, set to this
            out_type: cast return as this type

        Returns:
            Formatted value of type out_type

        """
        if value is None:
            value = default

        if isinstance(value, SpecialTagDirective):
            result = value.get_value(self)
            return types.cast_to_type(result, out_type)
        if isinstance(value, str):
            result = self.formatter.vformat(value, None, self)
            result_type = type(result)
            if out_type is result_type:
                # no need to cast, result is already desired type.
                return result
            elif out_type is bool:
                # casting a str to bool is always True, hence special case. If
                # the str value is 'False'/'false', presumably user can
                # reasonably expect a bool False response.
                return types.cast_to_bool(result)
            else:
                return out_type(result)
        else:
            return out_type(value)

    def get_formatted_value(self, input_value):
        """Run token substitution on the input against context.

        If input_value is a formattable string or SpecialTagDirective,
        will return the formatted result.

        If input_value is an iterable, will iterate input recursively and
        format all formattable objects it finds. Mappings will format both key
        and value.

        If input_value is not a string with a formatting expression such as
        'mystring{expr}morestring' and not iterable, will just return the input
        object. E.g An int input will return the same int.

        Choosing between get_formatted() and get_formatted_value():
        - get_formatted gets a context[key] value with formatting applied.
        - get_formatted_value is for any input object.

        An input string with a single formatting expression and nothing else
        will return the object at that context path:
        input_value = '{key1}'.

        This means that the return obj will be the same type as the source
        object. This return object in itself has token substitions run on it
        iteratively.

        By comparison, multiple formatting expressions and/or the inclusion
        of literal text will result in a string return type:
        input_value = '{key1} literal text {key2}'

        Then this will return string: "Piping down the valleys wild"

        This is not a full on deepcopy, and it's on purpose not a full on
        deepcopy. It will handle dict, list, set, tuple for iteration, without
        any especial cuteness for other types or types not derived from these.

        Returns:
            Iterable identical in structure to the input iterable, except
            where formatting changed a value from a string to the
            formatting expression's evaluated value.

        Args:
            input_value: Any object to format.

        Returns:
            any given type: Formatted value with {substitutions} made from
            context. If input was not a string, will just return input_value
            untouched.

        """
        return self.formatter.vformat(input_value, None, self)

    def get_processed_string(self, input_string):
        """Use get_formatted_value(input_value) instead. Deprecated."""
        from warnings import warn
        warn(
            ("Use get_formatted_value(input_value) instead of "
             "get_processed_string"),
            DeprecationWarning)
        return self.formatter.vformat(input_string, None, self)

    def iter_formatted_strings(self, iterable_strings):
        """Yield a formatted string from iterable_strings.

        If iterable_strings[0] = 'Piping {key1} the {key2} wild'
        And context = {'key1': 'down', 'key2': 'valleys', 'key3': 'value3'}

        Then the 1st yield is: "Piping down the valleys wild"

        Args:
            iterable: Iterable containing strings. E.g a file-like object.

        Returns:
            Yields formatted line.

        """
        for string in iterable_strings:
            yield self.formatter.vformat(string, None, self)

    def keys_exist(self, *keys):
        """Check if keys exist in context.

        Args:
            *keys: *args of str for keys to check in context.

        Returns:
            tuple(bool) where bool indicates the key does exist in context,
            same order as *keys.

        Sample:
            k1, = context.keys_exist('k1')
            k1, k2, k3 = context.keys_exist('k1', 'k2', 'k3')

        """
        return tuple(key in self.keys() for key in keys)

    def keys_of_type_exist(self, *keys):
        """Check if keys exist in context and if types are as expected.

        Args:
            *keys: *args for keys to check in context.
                   Each arg is a tuple(str, type)

        Returns:
            Tuple of namedtuple ContextItemInfo, same order as *keys.
            ContextItemInfo(key,
                            key_in_context,
                            expected_type,
                            is_expected_type)

            Remember if there is only one key in keys, the return assignment
            needs an extra comma to remind python that it's a tuple:
            # one
            a, = context.keys_of_type_exist('a')
            # > 1
            a, b = context.keys_of_type_exist('a', 'b')

        """
        # k[0] = key name, k[1] = exists, k2 = expected type
        keys_exist = [(key, key in self.keys(), expected_type)
                      for key, expected_type in keys]

        return tuple(ContextItemInfo(
            key=k[0],
            key_in_context=k[1],
            expected_type=k[2],
            is_expected_type=isinstance(self[k[0]], k[2])
            if k[1] else None,
            has_value=k[1] and not self[k[0]] is None
        ) for k in keys_exist)

    def merge(self, add_me):
        """Merge add_me into context and applies interpolation.

        Bottom-up merge where add_me merges into context. Applies string
        interpolation where the type is a string. Where a key exists in
        context already, add_me's value will overwrite what's in context
        already.

        Supports nested hierarchy. add_me can contains dicts/lists/enumerables
        that contain other enumerables et. It doesn't restrict levels of
        nesting, so if you really want to go crazy with the levels you can, but
        you might blow your stack.

        If something from add_me exists in context already, but add_me's value
        is of a different type, add_me will overwrite context. Do note this.
        i.e if you had context['int_key'] == 1 and
        add_me['int_key'] == 'clearly not a number', the end result would be
        context['int_key'] == 'clearly not a number'

        If add_me contains lists/sets/tuples, this merges these
        additively, meaning it appends values from add_me to the existing
        sequence.

        Args:
            add_me: dict. Merge this dict into context.

        Returns:
            None. All operations mutate this instance of context.

        """
        def merge_recurse(current, add_me):
            """Walk the current context tree in recursive inner function.

            On 1st iteration, current = self(i.e root of context)
            On subsequent recursive iterations, current is wherever you're at
            in the nested context hierarchy.

            Args:
                current: dict. Destination of merge.
                add_me: dict. Merge this to current.
            """
            for k, v in add_me.items():
                # key supports interpolation
                k = self.get_formatted_value(k)

                # str not mergable, so it doesn't matter if it exists in dest
                if isinstance(v, (str, SpecialTagDirective)):
                    # just overwrite dest - str adds/edits indiscriminately
                    current[k] = self.get_formatted_value(v)
                elif isinstance(v, (bytes, bytearray)):
                    # bytes aren't mergable or formattable
                    # only here to prevent the elif on enumerables catching it
                    current[k] = v
                # deal with things that are mergable - exists already in dest
                elif k in current:
                    if types.are_all_this_type(Mapping, current[k], v):
                        # it's dict-y, thus recurse through it to merge since
                        # it exists in dest
                        merge_recurse(current[k], v)
                    elif types.are_all_this_type(list, current[k], v):
                        # it's list-y. Extend mutates existing list since it
                        # exists in dest
                        current[k].extend(
                            self.get_formatted_value(v))
                    elif types.are_all_this_type(tuple, current[k], v):
                        # concatenate tuples
                        current[k] = (
                            current[k] + self.get_formatted_value(v))
                    elif types.are_all_this_type(Set, current[k], v):
                        # join sets
                        current[k] = (
                            current[k] | self.get_formatted_value(v))
                    else:
                        # at this point it's not mergable
                        current[k] = self.get_formatted_value(v)
                else:
                    # at this point it's not mergable, nor in context
                    current[k] = self.get_formatted_value(v)

        # first iteration starts at context dict root
        merge_recurse(self, add_me)

    def set_defaults(self, defaults):
        """Set defaults in context if keys do not exist already.

        Adds the input dict(defaults) into the context, only where keys in
        defaults do not already exist in context. Supports nested hierarchies.

        Example:
        Given a context like this:
            key1: value1
            key2:
                key2.1: value2.1
            key3: None

        And defaults input like this:
            key1: 'updated value here won't overwrite since it already exists'
            key2:
                key2.2: value2.2
            key3: 'key 3 exists so I won't overwrite

        Will result in context:
            key1: value1
            key2:
                key2.1: value2.1
                key2.2: value2.2
            key3: None

        Args:
            defaults: dict. Add this dict into context.

        Returns:
            None. All operations mutate this instance of context.

        """
        def defaults_recurse(current, defaults):
            """Walk the current context tree in recursive inner function.

            On 1st iteration, current = self(i.e root of context)
            On subsequent recursive iterations, current is wherever you're at
            in the nested context hierarchy.

            Args:
                current: dict. Destination of merge.
                defaults: dict. Add this to current if keys don't exist
                                already.

            """
            for k, v in defaults.items():
                # key supports interpolation
                k = self.get_formatted_value(k)

                if k in current:
                    if types.are_all_this_type(Mapping, current[k], v):
                        # it's dict-y, thus recurse through it to check if it
                        # contains child items that don't exist in dest
                        defaults_recurse(current[k], v)
                else:
                    # since it's not in context already, add the default
                    current[k] = self.get_formatted_value(v)

        # first iteration starts at context dict root
        defaults_recurse(self, defaults)
